Pick the data provider arguments that can be used to answer the following query.

=== Query ===
{{ user_query }}

=== Data Provider Documentation ===
{{ data_provider_documentation }}

=== Data Description ===
{{ data_description }}

=== Task ===
Print a valid json message with the following fields:
"feasible" - boolean "True" if the data provider can retrieve data for the user query and "False" otherwise
"justification" - string with feasibility justification
"arguments" - a json object with one field for each data provider argument with the right value
